v.setCursorPosition(2,8);
v.type("private:");
v.enter();
v.type("void ok();");
v.enter();
v.type("public:");
v.enter();
v.type("void ok();");
v.enter();
v.type("protected:");
v.enter();
v.type("void ok();");
