v.setCursorPosition(6, document.lineLength(6));
v.enter();
v.type("else");
v.enter();
v.type("nothing");
v.enter();
v.type("end");
v.enter();
v.type("print(i)")
v.enter();
v.type("print(i^2)");
v.enter();
v.type("end");
v.enter();
v.type("end");
