v.setCursorPosition(1,0);
v.enter();
v.type("class MyClass");
v.enter();
v.enter();
v.type("attr_reader :foo");
v.enter();
v.enter();
v.type("private");
v.enter();
v.enter();
v.type("def helper(str)");
v.enter();
v.type("puts \"helper(#{str})\"");
v.enter();
v.type("end");
v.enter();
v.enter();
v.type("end");
